import sys
import logging

from typing import Callable, Any
from timeit import default_timer as timer

logger = logging.getLogger('howfast_apm')


class Interaction:
    """ An external interaction with other services """
    # Can be "request"
    interaction_type: str
    # Name holds the URL if interaction_type is "request"
    name: str
    elapsed: float
    extra: dict

    def __init__(self, interaction_type, name, elapsed, extra=None):
        self.interaction_type = interaction_type
        self.name = name
        self.elapsed = elapsed
        self.extra = extra or {}

    def serialize(self):
        """ JSON-serialize the Interaction """
        return {
            "interaction_type": self.interaction_type,
            "name": self.name,
            "elapsed": self.elapsed,
            "extra": self.extra,
        }


def install_hooks(record_interaction: Callable[[Interaction], Any]) -> None:
    """ Install the HTTP hooks """
    patch_requests_module = True
    try:
        # Try to import the module to see if it's available
        import requests  # noqa
    except ModuleNotFoundError:
        # Maybe requests is not installed / available in the instrumented code
        patch_requests_module = False

    tmp_urllib = sys.modules['urllib']
    if patch_requests_module:
        tmp_requests = sys.modules['requests']

    def get_patched(func, meta_extractor: Callable):

        def patched_request(*args, **kwargs):
            start = timer()
            resp = func(*args, **kwargs)
            elapsed = timer() - start

            try:
                method, name = meta_extractor(*args, **kwargs)

                record_interaction(
                    Interaction(
                        interaction_type="request",
                        name=name,
                        elapsed=elapsed,
                        extra={'method': method.lower()},
                    ))
            # Catch any exception because we don't want it to bubble up to the real app
            except Exception:
                logger.error("Unable to record interaction", exc_info=True)  # pragma: nocover

            return resp

        return patched_request

    # TODO: build method extractor for urlopen
    tmp_urllib.request.urlopen = get_patched(
        tmp_urllib.request.urlopen,
        lambda *args, **kwargs: [None, None],
    )
    sys.modules['urllib'] = tmp_urllib

    if patch_requests_module:
        tmp_requests.request = get_patched(
            tmp_requests.request,
            lambda *args, **kwargs: [
                args[0] if len(args) > 0 else kwargs.get('method'),
                args[1] if len(args) > 1 else kwargs.get('url'),
            ],
        )

        def request_alias_extractor(method):
            return lambda *args, **kwargs: [
                method,
                args[0] if len(args) > 0 else kwargs.get('url'),
            ]

        tmp_requests.get = get_patched(
            tmp_requests.get,
            meta_extractor=request_alias_extractor('get'),
        )
        tmp_requests.post = get_patched(
            tmp_requests.post,
            meta_extractor=request_alias_extractor('post'),
        )
        tmp_requests.head = get_patched(
            tmp_requests.head,
            meta_extractor=request_alias_extractor('head'),
        )
        tmp_requests.put = get_patched(
            tmp_requests.put,
            meta_extractor=request_alias_extractor('put'),
        )
        tmp_requests.delete = get_patched(
            tmp_requests.delete,
            meta_extractor=request_alias_extractor('delete'),
        )

        sys.modules['requests'] = tmp_requests



import os
import logging
from typing import Optional, List
from datetime import datetime
from queue import Full, Empty

from .config import HOWFAST_APM_RECORD_INTERACTIONS
from .queue import queue
from .runner import Runner
from .hook_requests import install_hooks, Interaction

logger = logging.getLogger('howfast_apm')


class CoreAPM:
    """
    Base class that provides the shared code for:
    * starting the background thread
    * pushing points to the queue
    * storing external interactions
    """

    app_id: Optional[str]

    runner: Optional[Runner]

    record_interactions: bool
    interactions: List[Interaction] = []

    def __init__(self, record_interactions=HOWFAST_APM_RECORD_INTERACTIONS):
        self.record_interactions = bool(record_interactions)
        logger.debug("Interactions will %s", 'be enabled' if self.record_interactions else 'NOT be enabled')
        self.interactions = []

    def setup(
            self,
            # The HowFast app ID to use
            app_id: str = None,
    ):
        if app_id:
            self.app_id = app_id
        else:
            self.app_id = os.environ.get('HOWFAST_APM_DSN')

        if self.app_id:
            logger.info(f"HowFast APM configured with DSN {self.app_id}")
            self.start_background_thread()
            if self.record_interactions:
                self.setup_hooks()
        else:
            logger.warning("HowFast APM initialized with no DSN, reporting will be disabled.")

    def start_background_thread(self):
        """ Start the thread that will consume points from the queue and send them to the API """
        self.runner = Runner(queue=queue, app_id=self.app_id)
        self.runner.start()
        # TODO: stop the thread at some point?

    def setup_hooks(self) -> None:
        """ Install hooks to register what is slow """
        install_hooks(self.record_interaction)

    def record_interaction(self, interaction: Interaction) -> None:
        """ Save the interaction """
        self.interactions.append(interaction)

    def reset_interactions(self):
        self.interactions = []

    def save_point(
            self,
            time_request_started: datetime,
            time_elapsed: float,  # seconds
            method: str,
            uri: str,
            response_status: str = None,  # HTTP response status (200 OK, etc)
            endpoint_name: str = None,  # function name handling the request
            url_rule: str = None,  # Route pattern matched for this endpoint (/pet/<int:id>)
            is_not_found: bool = None,  # If the request did not match any route
    ) -> None:
        """
        Save a request/response performance information.

        This method is called by subclasses with their framework-specific information. We then add
        the core-level collected performance data (interactions) and call self._save_point().
        """
        self._save_point(
            time_request_started=time_request_started,
            time_elapsed=time_elapsed,
            method=method,
            uri=uri,
            interactions=self.interactions,
            response_status=response_status,
            endpoint_name=endpoint_name,
            url_rule=url_rule,
            is_not_found=is_not_found,
        )
        # Reset the list of interactions, since it's specific to a request/point
        self.reset_interactions()

    @staticmethod
    def _save_point(**kwargs) -> None:
        """ Save a request/response performance information """

        interaction_list = []
        interactions = kwargs.get('interactions', [])
        while interactions:
            interaction_list.append(interactions.pop().serialize())

        # Forward the arguments to the queue
        item = kwargs

        # Capped queue
        pushed = False
        while pushed is False:
            try:
                queue.put_nowait(item)
                pushed = True
            except Full:
                # The queue is full - let's pop the oldest element and discard it, and try again
                try:
                    queue.get_nowait()
                except Empty:
                    pass



import fnmatch
import re

from typing import List


def compile_endpoints(*blacklist):
    """Compiles a list endpoints to a list of regexes"""
    return [re.compile(fnmatch.translate(x)) for x in blacklist]


def is_in_blacklist(uri: str, blacklist: List[str]) -> bool:
    """
    Return True if the URI is blacklisted.

      >>> is_in_blacklist('/some-real-uri/', compile_endpoints('/exact-uri/'))
      False
      >>> is_in_blacklist('/exact-uri/', compile_endpoints('/exact-uri/'))
      True
      >>> is_in_blacklist('/job/42/retry', compile_endpoints('/job/*'))
      True

    """
    for pattern in blacklist:
        if pattern.match(uri):
            return True
    return False


if __name__ == "__main__":
    import doctest
    doctest.testmod()



import logging
from typing import List
from datetime import datetime, timezone
from timeit import default_timer as timer
from flask.signals import request_started
from flask import Flask, request
from werkzeug import local, exceptions

from .core import CoreAPM
from .utils import is_in_blacklist, compile_endpoints

logger = logging.getLogger('howfast_apm')


class HowFastFlaskMiddleware(CoreAPM):
    """
    Flask middleware to measure how much time is spent per endpoint.

    This implementation is purposedly naive and potentially slow, but its goal is to validate the
    PoC. It should be replaced/improved in the future, based on the results of the PoC.
    """

    def __init__(
            self,
            # The Flask application to analyze
            app: Flask,
            # The HowFast app ID to use
            app_id: str = None,
            # Endpoints not to monitor
            endpoints_blacklist: List[str] = None,
            # Other configuration parameters passed to the CoreAPM constructor
            **kwargs,
    ):
        super().__init__(**kwargs)

        self.app = app
        self.wsgi_app = app.wsgi_app

        # We need to store thread local information, let's use Werkzeug's context locals
        # (see https://werkzeug.palletsprojects.com/en/1.0.x/local/)
        self.local = local.Local()
        self.local_manager = local.LocalManager([self.local])

        # Overwrite the passed WSGI application
        app.wsgi_app = self.local_manager.make_middleware(self)

        if endpoints_blacklist:
            self.endpoints_blacklist = compile_endpoints(*endpoints_blacklist)
        else:
            self.endpoints_blacklist = []

        # Setup the queue and the background thread
        self.setup(app_id)

        request_started.connect(self._request_started)

    def __call__(self, environ, start_response):
        if not self.app_id:
            # HF APM not configured, return early to save some time
            # TODO: wouldn't it be better to just not replace the WSGI app?
            return self.wsgi_app(environ, start_response)

        uri = environ.get('PATH_INFO')

        if is_in_blacklist(uri, self.endpoints_blacklist):
            # Endpoint blacklist, return now
            return self.wsgi_app(environ, start_response)

        method = environ.get('REQUEST_METHOD')

        response_status: str = None

        def _start_response_wrapped(status, *args, **kwargs):
            nonlocal response_status
            # We wrap the start_response callback to access the response status line (eg "200 OK")
            response_status = status
            return start_response(status, *args, **kwargs)

        time_request_started = datetime.now(timezone.utc)

        try:
            # Time the function execution
            start = timer()
            return_value = self.wsgi_app(environ, _start_response_wrapped)
            # Stop the timer as soon as possible to get the best measure of the function's execution time
            end = timer()
        except BaseException:
            # The WSGI app raised an exception, let's still save the point before raising the
            # exception again
            # First, "stop" the timer now to get the good measure of the function's execution time
            end = timer()
            # The real response status will actually be set by the server that interacts with the
            # WSGI app, but we cannot instrument it from here, so we just assume a common string.
            response_status = "500 INTERNAL SERVER ERROR"
            raise
        finally:
            elapsed = end - start

            self.save_point(
                time_request_started=time_request_started,
                time_elapsed=elapsed,
                method=method,
                uri=uri,
                response_status=response_status,
                # Request metadata
                endpoint_name=getattr(self.local, 'endpoint_name', None),
                url_rule=getattr(self.local, 'url_rule', None),
                is_not_found=getattr(self.local, 'is_not_found', None),
            )
            # TODO: remove this once overhead has been measured in production
            logger.info("overhead when saving the point: %.3fms", (timer() - end) * 1000)

        return return_value

    def _request_started(self, sender, **kwargs):
        with sender.app_context():
            self._save_request_metadata()

    def _save_request_metadata(self):
        """ Extract and save request metadata in the context local """
        # This will yield strings like:
        # * "monitor" (when the endpoint is defined using a resource)
        # * "apm-collection.store_points" (when the endpoint is defined with a blueprint)
        # The endpoint name will always be lowercase
        self.local.endpoint_name = request.endpoint
        # This will yield strings like "/v1.1/apm/<int:apm_id>/endpoint"
        self.local.url_rule = request.url_rule.rule if request.url_rule is not None else None
        # We want to tell the difference between a "real" 404 and a 404 returned by an existing view
        self.local.is_not_found = isinstance(request.routing_exception, exceptions.NotFound)



import os

# Where to send the performance data
HOWFAST_APM_COLLECTOR_URL = os.environ.get(
    'HOWFAST_APM_COLLECTOR_URL',
    "https://api.howfast.tech/v1.1/apm-collector/store",
)

# Record interactions of the API with external sources (HTTP requests, etc)
HOWFAST_APM_RECORD_INTERACTIONS = os.environ.get(
    'HOWFAST_APM_RECORD_INTERACTIONS',
    False,
)



import requests
from logging import getLogger
from threading import Thread
from typing import List, Dict, Any

from queue import Queue, Empty
from .config import HOWFAST_APM_COLLECTOR_URL

logger = getLogger("howfast_apm")


class Runner(Thread):
    """ Thread dedicated to sending performance events stored in the queue to the API """

    # The DSN of the application
    app_id: str

    # If the queue is empty, wait up to X seconds before checking if the thread has to stop.
    # Whatever this value, an incoming element in the queue will be picked up as soon as it arrives.
    sleep_delay = 0.5

    # Group points before sending them to the API
    batch_size = 100

    # Local list of the points to be sent to the API
    current_batch: List[Dict[str, Any]]

    def __init__(self, queue: Queue, app_id: str):
        self.queue = queue
        self.app_id = app_id
        self.current_batch = []
        # TODO: stop mechanism?
        self.stop = False
        logger.debug("APM thread starting...")
        super(Runner, self).__init__(
            name="HowFast APM",
            # The entire Python program exits when no alive non-daemon threads are left, and we
            # don't want this thread to block the program from exiting.
            # TODO: block until the queue is empty?
            # TODO: don't abruptly kill this thread when the process stops, but instead set the
            # `stop` property
            daemon=True,
        )

    def run(self):
        while self.stop is False:
            self.run_once()

    def run_once(self):
        try:
            for _ in range(self.batch_size):
                # Try to get N=100 points from the queue. As soon as the queue is empty, queue.get
                # will block up to self.sleep_delay (0.5s). Once the queue is empty after the sleep
                # delay OR 10 points have been retrieved from the queue, we proceed to sending the
                # batch. This strategy means that the thread will wait up to N=100 * 0.5s = 50s
                # before sending points to the API, if 100 points arrive with slightly less than
                # 0.5s between them.
                point = self.queue.get(timeout=self.sleep_delay)
                self.current_batch.append(point)

        except Empty:
            pass
        except Exception:
            logger.error("Runner crashed:", exc_info=True)
            return

        if self.current_batch:
            # If the queue was empty, the current batch will be empty and we don't need to send the batch
            self._send_batch_robust()

        # Exit now if should stop
        if self.stop:
            return

    @staticmethod
    def serialize_point(point: dict) -> dict:
        """ Prepare the point to be sent to the API """
        serialized_point = {
            'method': point['method'],
            'uri': point['uri'],
            'time_request_started': point['time_request_started'].isoformat(),
            'time_elapsed': point['time_elapsed'],
            'interactions': point['interactions'],
            'response_status': point['response_status'],
            'endpoint_name': point['endpoint_name'],
        }
        # Save some space in the request body if we don't have interesting information
        if 'url_rule' in point and point['url_rule']:
            serialized_point['url_rule'] = point['url_rule']
        if 'is_not_found' in point and point['is_not_found'] is not None:
            serialized_point['is_not_found'] = point['is_not_found']

        return serialized_point

    def _send_batch_robust(self, attempts=1, max_attempts=3) -> None:
        """ Retry sending the batch up to max_retry times """
        try:
            self.send_batch()
        except Exception:
            # Print an error, and don't die
            logger.error(
                "Runner was unable to send performance data (try %d/%d)",
                attempts,
                max_attempts,
                exc_info=True,
            )
            if attempts < max_attempts:
                # Retry sending the batch
                self._send_batch_robust(
                    attempts=attempts + 1,
                    max_attempts=max_attempts,
                )
            else:
                # We've retried enough times, let's just drop the batch :(
                logger.warning(
                    "Unable to send the batch after %d attemps, dropping %d points",
                    max_attempts,
                    len(self.current_batch),
                )
                # Drop the points
                self.current_batch = []

    def send_batch(self) -> None:
        """ Process one performance point """
        logger.debug("Posting %d point(s) to the server", len(self.current_batch))
        response = requests.post(
            HOWFAST_APM_COLLECTOR_URL,
            json={
                'dsn': self.app_id,
                'perf': list(map(self.serialize_point, self.current_batch)),
            },
        )
        # Batch is now empty
        self.current_batch = []

        if response.status_code != 200:
            logger.warning(
                "Unable to send a point to the server (%s), data was dropped! %s",
                response.status_code,
                response.content,
            )
